## ------------------------------------------------------------------------
setClass(Class = "Rohstoffe",
         slots = c(laenge = "numeric", ## Längengrad +/- 180 (ost: +, west: -)
                   breite = "numeric", ## Breitengrand +/- 90 (nord: +, sued: -)
                   tiefe = "numeric",  ## Tiefe in Meter
                   menge = "numeric",  ## Menge
                   einheit = "character", ## Einheit der Menge
                   art = "character")) ## Art/Name des Rohstoffes

## ------------------------------------------------------------------------
isClass("Rohstoffe")

## ------------------------------------------------------------------------
getClass("Rohstoffe")
getClassDef("Rohstoffe")

## ------------------------------------------------------------------------
R <- new("Rohstoffe", laenge = 11.69, breite = 50.01,
         tiefe = 100, ## in der Nähe von Goldkronach (Oberfranken)
         menge = 0.1, einheit = "kg", art = "Gold")

## ------------------------------------------------------------------------
is(R, "Rohstoffe")

## ------------------------------------------------------------------------
R
show(R)
## aber auch möglich
print(R)

## ------------------------------------------------------------------------
new("Rohstoffe")

## ------------------------------------------------------------------------
getSlots("Rohstoffe")
slotNames("Rohstoffe")

## ------------------------------------------------------------------------
R@laenge
R@lae ## nein!
## oder auch
slot(R, "breite")
slot(R, "br")

## ------------------------------------------------------------------------
R@laenge <- 12
R
## oder auch
slot(R, "laenge") <- 11
R

## ------------------------------------------------------------------------
removeClass("Rohstoffe")
isClass("Rohstoffe") ## nicht mehr vorhanden

## ------------------------------------------------------------------------
setClass(Class = "Position", slots = c(laenge = "numeric",
                                       breite = "numeric",
                                       tiefe = "numeric"),
         ## Default Prototyp
         prototype = prototype(laenge = 0, breite = 0, tiefe = 0),
         ## validity check
         validity = function(object){
            if(any(abs(object@laenge) > 180))
              stop("Ungültiger Längengrad!")
            if(any(abs(object@breite) > 90))
              stop("Ungültiger Breitengrad!")
            if(any(object@tiefe < 0) | any(object@tiefe > 6350000))
              stop("Ungültige Tiefe!")
            n <- length(object@laenge)
            if(length(object@breite) != n | length(object@tiefe) != n)
              stop("laenge, breite und tiefe besitzen unterschiedliche Längen")
            return(TRUE)
         })

## ------------------------------------------------------------------------
new("Position")
new("Position", tiefe = 10)
new("Position", tiefe = -10) ## fehlerhafte Eingabe
new("Position", laenge = 190) ## fehlerhafte Eingabe
new("Position", breite = -100) ## fehlerhafte Eingabe
new("Position", laenge = 11.69, breite = 50.01, tiefe = 100)

## ------------------------------------------------------------------------
setClass(Class = "Rohstoffe",
         slots = c(menge = "numeric",
                   einheit = "character",
                   art = "character"),
         contains = "Position",
         prototype = prototype(menge = 0, einheit = "kg", art = ""),
         validity = function(object){
            if(any(object@menge < 0))
              stop("menge darf nicht negativ sein!")
            if(!(object@einheit %in% c("mg", "g", "kg", "t")))
              stop("Es wurde keine korrekte Mengeneinheit angegeben.")
            n <- length(object@laenge)
            if(length(object@menge) != n | length(object@art) != n)
              stop("laenge, breite, tiefe und menge bzw. art
                    sind nicht von gleicher Länge")
            TRUE
         })
getClass("Rohstoffe")

## ------------------------------------------------------------------------
getValidity(getClassDef("Rohstoffe"))
## neuer validity-Check
validRohstoffe <- function(object){
    if(any(object@menge < 0))
      stop("menge darf nicht negativ sein!")
    if(!(object@einheit %in% c("mg", "g", "kg", "t")))
      stop("Es wurde keine korrekte Mengeneinheit angegeben.")
    n <- length(object@laenge)
    if(length(object@menge) != n | length(object@art) != n)
      stop("laenge, breite, tiefe und menge bzw. art
            sind nicht von gleicher Laenge")
    if(object@einheit == "kg" & sum(object@menge) > 6e24)
      stop("Die eingegebenen Mengen sind größer als die Erdmasse!")
    TRUE
}
setValidity("Rohstoffe", validRohstoffe)
getValidity(getClassDef("Rohstoffe"))

## ------------------------------------------------------------------------
extends("Rohstoffe", "Position")

## ------------------------------------------------------------------------
new("Rohstoffe")
new("Rohstoffe", laenge = 11.69, breite = 50.01, tiefe = 100,
                 menge = 0.1, einheit = "kg", art = "Gold")

## ------------------------------------------------------------------------
## generating function für Klasse Position
Position <- function(laenge, breite, tiefe){
  new("Position", laenge = laenge, breite = breite, tiefe = tiefe)
}
P <- Position(laenge = 11.69, breite = 50.01, tiefe = 10)
P
## generating function für Klasse Rohstoffe
Rohstoffe <- function(Pos, laenge, breite, tiefe, menge, einheit, art){
  if(!missing(Pos)){
    laenge <- Pos@laenge
    breite <- Pos@breite
    tiefe <- Pos@tiefe
  }
  new("Rohstoffe", laenge = laenge, breite = breite, tiefe = tiefe,
                   menge = menge, einheit = einheit, art = art)
}
R <- Rohstoffe(Pos = P, menge = 0.1, einheit = "kg", art = "Gold")
R
is(R, "Rohstoffe")
is(R, "Position") ## TRUE, da Position Mutterklasse von Rohstoffe

## ------------------------------------------------------------------------
## Die Slots "menge", "einheit" und "art" werden hinzugefügt und mit
## Default-Werten aus dem Prototyp belegt.
as(P, "Rohstoffe")
## Die Slots "menge", "einheit" und "art" werden entfernt.
as(R, "Position")

## ------------------------------------------------------------------------
isClass("numeric")
isClass("data.frame")
## oder
getClassDef("numeric")
getClass("lm") ## virtuelle Klasse!

## ------------------------------------------------------------------------
isClass("estimate")
## Wir machen hieraus eine formale S4 Klasse
setOldClass("estimate")
isClass("estimate")
getClassDef("estimate")

## ------------------------------------------------------------------------
## Vererbung
extends("integer", "vector")
extends("integer", "numeric")

## ------------------------------------------------------------------------
getClass("vector")
## Konkret sind dies dann Zahlenvektoren
getClass("numeric")
## oder Zeichenvektoren
getClass("character")
## oder auch Listen
getClass("list")

## ------------------------------------------------------------------------
setClass("Vektor", slots = c(data = "ANY"),
                   contains = "VIRTUAL")
getClass("Vektor")
## Für virtuelle Klassen können keine Objekte erzeugt werden
new("Vektor")

## ------------------------------------------------------------------------
setClass("reell", contains = "Vektor",
         prototype = prototype(data = numeric(0)),
         validity = function(object){
            if(!is.numeric(object@data))
                stop("data enthält keinen Vektor von reellen Zahlen!")
            return(TRUE)
         })
new("reell")

## ------------------------------------------------------------------------
getClassDef("OptionalFunction")

## ------------------------------------------------------------------------
setClassUnion("OptionalNumeric", c("NULL", "numeric"))
getClassDef("OptionalNumeric")

## ------------------------------------------------------------------------
library(microbenchmark)
myFun <- function(R, laenge, breite, tiefe, menge, einheit, art){
    new("Rohstoffe", laenge = laenge, breite = breite, tiefe = tiefe,
                     menge = menge, einheit = einheit, art = art)
}
myFun1 <- function(R, laenge, breite, tiefe, menge, einheit, art){
    R@laenge <- laenge
    R@breite <- breite
    R@tiefe <- tiefe
    R@menge <- menge
    R@einheit <- einheit
    R@art <- art
    R
}
R
res <- microbenchmark(
  myFun(R = R, laenge = 11.5, breite = 50, tiefe = 10, menge = 1,
        einheit = "kg", art = "Gold"),
  myFun1(R = R, laenge = 11.5, breite = 50, tiefe = 10, menge = 1,
         einheit = "kg", art = "Gold"),
  times = 10000, unit = "relative")
levels(res$expr) <- c("mit neuem Objekt", "kein neues Objekt")
## relative Unterschiede bezogen auf das schnellste Verfahren
summary(res)[,1:7]
autoplot(res)

## ------------------------------------------------------------------------
getGenerics()

## ------------------------------------------------------------------------
isGeneric("print")
isGeneric("plot")
isGeneric("initialize")

## ------------------------------------------------------------------------
new ## -> ruft initialize auf
args("initialize")
getGeneric("initialize")
findMethodSignatures("initialize")

## ------------------------------------------------------------------------
setMethod(f = "initialize", signature = c("Rohstoffe"),
          definition = function(.Object, Pos, laenge, breite, tiefe, menge,
                                einheit, art){
                          if(!missing(Pos)){
                              laenge <- Pos@laenge
                              breite <- Pos@breite
                              tiefe <- Pos@tiefe
                          }
                          .Object@laenge <- laenge
                          .Object@breite <- breite
                          .Object@tiefe <- tiefe
                          .Object@menge <- menge
                          .Object@einheit <- einheit
                          .Object@art <- art

                          .Object
                       })
new("Rohstoffe", Pos = P, menge = 0.1, einheit = "kg", art = "Gold")

## ------------------------------------------------------------------------
showMethods("initialize")
## oder nur die Signaturen
findMethodSignatures("initialize")

## ------------------------------------------------------------------------
existsMethod("initialize", signature = "Rohstoffe")
hasMethod("initialize", signature = "Rohstoffe")

## ------------------------------------------------------------------------
hasMethods("initialize")

## ------------------------------------------------------------------------
findMethods("initialize", classes = "Rohstoffe")
getMethod("initialize", signature = "Rohstoffe")
selectMethod("initialize", signature = "Rohstoffe")
showMethods("initialize", classes = "Rohstoffe", includeDefs = TRUE)
## Wo definiert?
findMethod("initialize", signature = "Rohstoffe")

## ------------------------------------------------------------------------
removeMethod(f = "initialize", signature = "Rohstoffe")

## ------------------------------------------------------------------------
args("show")
getGeneric("show")
args("print")

## ------------------------------------------------------------------------
P ## entspricht show(P)
show(P)
print(P)
## oder z.B.
show(diag(3))

## ------------------------------------------------------------------------
showMethods("show")
## Die Default Methode
getMethod("show", "ANY")
## ruft die Funktion "showDefault" auf.
showDefault

## ------------------------------------------------------------------------
setMethod("show", signature = "Rohstoffe",
    definition = function(object){
        for(i in seq_len(length(object@art))){
            if(i > 1)
            cat("=========================================================\n")
            cat("Art des Rohstoffs:\t", object@art[i], "\n")
            cat("\nGeographische Position\n")
            cat("Längengrad:\t", object@laenge[i], "\n")
            cat("Breitengrad:\t", object@breite[i], "\n")
            cat("\nGeschätzte Tiefe des Vorkommens:\t", object@tiefe[i], "m\n")
            cat("\nGeschätzte Menge des Vorkommens:\t", object@menge[i],
                object@einheit, "\n")
        }
    })

## ------------------------------------------------------------------------
P1 <- Position(laenge = c(11.69, 10.9, 11.8), breite = c(50.01, 50.1, 49.8),
               tiefe = c(10, 50, 100))
P1 ## immer noch showDefault!
R1 <- Rohstoffe(P1, art = c("Gold", "Eisenerz", "Uran"),
                menge = c(0.1, 1000, 20), einheit = "kg")
R1

## ------------------------------------------------------------------------
getGeneric("plot")
args("plot")
setMethod("plot", signature = c("Rohstoffe", "ANY"),
## oder auch kürzer setMethod("plot", signature = "Rohstoffe"
## fehlende Argumente werden dann automatisch als "ANY" behandelt
    definition = function(x, y = NULL, ...){
        plot(x@breite, x@laenge, xlab = "Breitengrad", ylab = "Längengrad",
             xlim = range(x@breite) + c(-0.5, 0.5),
             ylim = range(x@laenge) + c(-0.5, 0.5), ...)
        text(x@breite, x@laenge, pos = 3,
             labels = paste(x@art, ": ", x@tiefe, "m, ", x@menge,
                            x@einheit, sep = ""))
    })

## ------------------------------------------------------------------------
plot(R1)

## ------------------------------------------------------------------------
setGeneric("addRohstoffe", ## Name
           function(obj, R, ...){ ## Argumente der generischen Funktion
             ## der Aufruf von standardGeneric sorgt für das Dispatching
             standardGeneric("addRohstoffe")
           })
## Wir definieren nun eine Methode für "addRohstoffe"
setMethod("addRohstoffe", signature = c("Rohstoffe", "Rohstoffe"),
    ## nicht notwendigerweise auch "..." in der Argumentliste
    definition = function(obj, R){
        obj@laenge <- c(obj@laenge, R@laenge)
        obj@breite <- c(obj@breite, R@breite)
        obj@tiefe <- c(obj@tiefe, R@tiefe)
        obj@art <- c(obj@art, R@art)
        obj@menge <- c(obj@menge, R@menge)
        obj
    })
addRohstoffe(R1, R)

## ------------------------------------------------------------------------
P
## eigentlich unzulässig, wird jedoch nicht automatisch geprüft!
P@tiefe <- -20
P
## Kommt erst zu Tage durch
validObject(P)

## ------------------------------------------------------------------------
## Generische Funktionen für die Accessor-Methoden
setGeneric("laenge", function(object) standardGeneric("laenge"))
setGeneric("breite", function(object) standardGeneric("breite"))
setGeneric("tiefe", function(object) standardGeneric("tiefe"))
## Generische Funktionen für die Replacement-Methoden
setGeneric("laenge<-", function(object, value) standardGeneric("laenge<-"))
setGeneric("breite<-", function(object, value) standardGeneric("breite<-"))
setGeneric("tiefe<-", function(object, value) standardGeneric("tiefe<-"))

## ------------------------------------------------------------------------
setMethod("laenge", signature = "Position",
          definition = function(object) object@laenge)
setMethod("breite", signature = "Position",
          definition = function(object) object@breite)
setMethod("tiefe", signature = "Position",
          definition = function(object) object@tiefe)
laenge(P1)
breite(P1)
tiefe(P1)

## ------------------------------------------------------------------------
## 1. Möglichkeit
setMethod("laenge<-", signature = c("Position", "numeric"),
      definition = function(object, value){
          if(any(abs(value) > 180))
              stop("Ungültiger Wert für den Längengrad")
          if(length(value) != length(object@laenge))
              stop("Falsche Länge des Längengradvektors")
          object@laenge <- value
          object
      })
P
laenge(P) <- 11.75
P
laenge(P) <- 190
P
removeMethod("laenge<-", signature = c("Position", "numeric"))
## 2. Kanonische Möglichkeit
setReplaceMethod("laenge", signature = c("Position", "numeric"),
      definition = function(object, value){
          if(any(abs(value) > 180))
              stop("Ungültiger Wert für den Längengrad")
          if(length(value) != length(object@laenge))
              stop("Falsche Länge des Längengradvektors")
          object@laenge <- value
          object
      })
P
laenge(P) <- 11.65
P
laenge(P) <- 190
P

## ------------------------------------------------------------------------
## Slot breite
setReplaceMethod("breite", signature = c("Position", "numeric"),
      definition = function(object, value){
          if(any(abs(value) > 90))
              stop("Ungültiger Wert für den Breitengrad")
          if(length(value) != length(object@breite))
              stop("Falsche Länge des Breitengradvektors")
          object@breite <- value
          object
      })
P
breite(P) <- 50.5
P
breite(P) <- -100
## Slot tiefe
setReplaceMethod("tiefe", signature = c("Position", "numeric"),
      definition = function(object, value){
          if(any(value < 0) | any(value > 6350000))
              stop("Ungültiger Wert für die Tiefe")
          if(length(value) != length(object@tiefe))
              stop("Falsche Länge des Tiefenvektors")
          object@tiefe <- value
          object
      })
P
tiefe(P) <- -10
tiefe(P) <- 50
P

## ------------------------------------------------------------------------
## 1. Definition der Klasse
setClass(Class = "MatrixList",
         prototype = prototype(list(new("matrix"))),
         contains = "list",
         validity = function(object){
            nrvalues <- length(object)
            for(i in 1:nrvalues)
              if(!is(object[[i]], "matrix"))
                 stop("Element ", i, " is keine Matrix")
            return(TRUE)
          })
## 2. Generating function
MatrixList <- function(...){
    new("MatrixList", list(...))
}

## ------------------------------------------------------------------------
getClassDef("MatrixList")
M <- MatrixList(diag(3), diag(2))
M
as(M, "list")

## ------------------------------------------------------------------------
setIs(class1 = "MatrixList", class2 = "matrix",
      test = function(obj) { length(obj) == 1 },
      coerce = function(obj) { obj[[1]] },
      replace = function(obj, value){ obj <- value[[1]] })
M
is(M, "matrix") ## nein
## aber
M1 <- MatrixList(diag(3))
M1
is(M1, "matrix") ## ja
as(M1, "matrix")

## ------------------------------------------------------------------------
setIs(class1 = "matrix", class2 = "MatrixList",
    coerce = function(obj){ new("MatrixList", list(obj)) },
    replace = function(obj, value){ obj[[1]] <- value })

## ------------------------------------------------------------------------
setAs(from = "matrix", to = "MatrixList",
    def = function(from){ new("MatrixList", list(from)) })
D <- diag(3)
is(D, "MatrixList") ## nein
## aber
as(D, "MatrixList")

## ------------------------------------------------------------------------
getGroupMembers("Arith")

## ------------------------------------------------------------------------
getGeneric("Arith")
setMethod("Arith", signature = c("MatrixList", "numeric"),
    definition = function(e1, e2){
        for(i in seq_len(length(e1))){
            e1[[i]] <- callGeneric(e1[[i]], e2)
        }
        e1
    })
M * 3
M / c(1, 2, 3)
M - 10

